#ifndef DEBUG
#define DEBUG
#endif

uint8_t init(int *fd, uint32_t if_index, uint8_t c, uint8_t p, uint8_t all_servers, uint8_t want_to_reconfigure){
  /*
   * This function is a part of library which applications 
   * will use.
   * Sending initialization message to client daemon on
   * behalf of the application. This would send
   * (1) msg_type   = 0
   * (2) pid        = process id of the application
   * (3) request_no = 1 because this is start of connection
   * (4) if_index   = interface requested by appl
   * (5) c          = deallocate all releasable resources
   * (6) p          = list of subnet prefixes.
   * (7) a          = want list of all servers
   * (8) w          = want to reconfigure
   *
   * Note that (1), (2) and (3) are part of header.
   *
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |   msg_type    | error(8 bits) |        pid (16 bits)          |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                   request_no (32 bits)                        |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                    if index (32 bits)                         |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |            Reserved (28 bits)                         |c|p|a|w|
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   *
   * On failure 0 is returned by function and fd remains unchanged.
   * On success 1 is returned and fd is set to socket descriptor
   * used for talking to client daemon.
   */
  SOLICIT_PARAMS         solicit;
  HEADER_CLIENTD_PACKET  header;
  struct iovec           iov[2];
  
  /* Check validity of parameters passed to this function */
  if(c > 1 || p > 1 || all_servers >1 || want_to_reconfigure > 1 || fd == NULL){
#ifdef DEBUG
    printf("Invalid parameters to init by the appl\n");
#endif
    return 0;
  }

  bzero((char *) &clientd_sockaddr, sizeof(clientd_sockaddr));

#ifdef SIN6_LEN
  clientd_sockaddr.sin6_len      = sizeof(clientd_sockaddr);      /* 4.3 BSD doesn't have this field      */
#endif

  clientd_sockaddr.sin6_family   = AF_INET6;                      /* Family as required for Ipv6          */
  clientd_sockaddr.sin6_port     = htons(CLD_TCP_PORT);           /* CLD listens on CLD_TCP_PORT          */
  clientd_sockaddr.sin6_flowinfo = 0;                             /* Flow & Traffic Class are unimportant */
  clientd_sockaddr.sin6_addr     = in6addr_loopback;              /* CLD is on same machine               */
  clientd_sockaddr.sin6_scope_id = 0;                             /* Not really required.                 */

  /* Open a TCP socket for the talking to client daemon   */
  sockfd         = socket(AF_INET6, SOCK_STREAM, 0);

  if(sockfd < 0){
    printf("Error in opening stream socket! Returning...\n");
    return 0;
  }

  /* Connect to the client daemon running on CLIENTD_PORT */

  if(connect(sockfd, (struct sockaddr *) &clientd_sockaddr, sizeof(clientd_sockaddr)) < 0){
    printf("Error in connecting to the client daemon. Returning...\n");
    return 0;
  }

  header.error      = SUCCESS;
  header.msg_type   = 0;
  header.pid        = htons(getpid());
  header.request_no = htonl(request_no++);
  
  solicit.if_index  = htonl(if_index);

  bzero(&solicit.res_c_p_a_w, sizeof(solicit.res_c_p_a_w));
  SET_BITS_MIDDLE(solicit.res_c_p_a_w, c, 4, 1);
  SET_BITS_MIDDLE(solicit.res_c_p_a_w, p, 3, 1);
  SET_BITS_MIDDLE(solicit.res_c_p_a_w, all_servers, 2, 1);
  SET_BITS_MIDDLE(solicit.res_c_p_a_w, want_to_reconfigure, 1, 1);
  solicit.res_c_p_a_w = htonl(solicit.res_c_p_a_w);

#ifdef DEBUG
  printf("Sending to client daemon: \nmsg_type = %u pid        = %u \nrequest_no = %u\n",
	 header.msg_type,
	 ntohs(header.pid),
	 ntohl(header.request_no));
  printf("if_index = %u \nc = %u p = %u all_servers = %uwant_to_reconfigure = %u\n",
	 ntohl(solicit.if_index),
	 GET_BITS_MIDDLE(ntohl(solicit.res_c_p_a_w), 4, 1),
	 GET_BITS_MIDDLE(ntohl(solicit.res_c_p_a_w), 3, 1),
	 GET_BITS_MIDDLE(ntohl(solicit.res_c_p_a_w), 2, 1),
	 GET_BITS_MIDDLE(ntohl(solicit.res_c_p_a_w), 1, 1));
#endif
  
  iov[0].iov_base   = (char *) &header;
  iov[0].iov_len    = sizeof(header);
  iov[1].iov_base   = (char *) &solicit;
  iov[1].iov_len    = sizeof(solicit);
  n                 = iov[0].iov_len + iov[1].iov_len;

  if(writev(sockfd, &iov[0], 2) != n){
    printf("Error in sending initialization message to client daemon. Returning...\n");
    return 0;
  }

#ifdef DEBUG
  printf("Waiting for reply to init msg\n");
#endif

  n = recv(sockfd, recv_msg, MAX_MSG_SIZE, 0);

  if(n < 0){
    printf("Error in recv. Returning...\n");
    return 0;
  }
  if(n == 0){
    printf("Connection closed by client daemon. Returning...\n");
    return 0;
  }

  memcpy(&header, recv_msg, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

#ifdef DEBUG
  printf("Packet-> ");
  print_packet(recv_msg, n);
  printf("\n");
  printf("Reply from client daemon\n\n");
  printf("msg_type   = %u\n", header.msg_type);
  printf("pid        = %u\n", header.pid);
  printf("request_no = %u\n", header.request_no);
#endif

  /*
   * Imp: Now, you need to process remaining part of message
   * to check if req was successful. If no, return 0. If yes,
   * send necessary data to appl.
   */
  *fd = sockfd;
  return 1;
}
