int request(struct in6_addr *unicast_addr, struct in6_addr *relay_addr,
	    struct in6_addr *server_addr, uint8_t C, uint8_t R,
	    uint8_t T, uint16_t transaction_id, IP_ADDR_EXT *ip_ext_head,
	    uint16_t *ere, int no_of_ere,
	    char *remaining_ext, int remaining_ext_length) {
  /*
   * This function is a part of library which applications 
   * will use.
   * Sending initialization message to client daemon on
   * behalf of the application. This would send
   *
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |   msg_type    | error(8 bits) |        pid (16 bits)          |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                   request_no (32 bits)                        |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |  Reserved (12 bits)   |U|C|R|T|   transaction_id (16 bits)    |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                                                               |
   * |       client's unicast address if present (16 octets)         |
   * |                                                               |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                                                               |
   * |                relay-address (16 octets)                      |
   * |                                                               |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                                                               |
   * |                server-address (16 octets)                     |
   * |                                                               |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |              extensions (variable number and length)          |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * Note: if T = 1, transaction id should be the one copied from
   *	   Reconfigure_init message. This is the job of appl to get
   *       the appropiate transaction id in such a case.
   *	   if T = 0, transaction id in above message MUST be zero.
   *       CLD would pick up a suitable transaction id.
   *
   *	   U = 1 means client's unicast address field is present
   *	   and CLD should not use relay for sending this message.
   *	   if U = 0, relay should be used.
   *
   * (1) msg_type   = 0
   * (2) pid        = process id of the application
   * (3) request_no = 1 because this is start of connection
   */

  /* If unicast addr is null library will not send unicast_addr
   * field in the request to client daemon
   * 
   * ip_ext_head         : pointer to head of list of ip adddress extensions to be sent.
   *                       Note the list MUST be terminated by a NULL pointer.
   * ere                 : Extension requests in the form of an array of "type of ext desired".
   * no_of_ere           : no of entries in above ere array. This field MUST be zero if there
   *                       are no extensions to ask for.
   * remainimg_ext       : extensions other than IP or ERE. These MUST be in format as given
   *                       in draft-ietf-dhc-v6exts-12.txt i.e. with type and length fields.
   *                       and in network byte order.
   * remaining_ext_length: total size of remaining extensions including the type, len fields.
   *                       This field MUST be zero if there are no extensions.
   */
  
  typedef struct type_len_settings {
    uint16_t type;
    uint16_t length;
    uint32_t settings;
    struct type_len_settings *next;
  } TYPE_LEN_SETTINGS;

  REQUEST_PARAMS         req_p;
  HEADER_CLIENTD_PACKET  header;
  struct iovec           iov[128];
  struct in6_addr        unspecified_addr;
  TYPE_LEN               ere_tl;
  TYPE_LEN_SETTINGS     *tls_head = NULL, **temp_tls = &tls_head, *t_tls;
  int i = 0, len = 0, count;

  /* Check validity of parameters passed to this function */
  if(C > 1 || R > 1 || T > 1){
#ifdef DEBUG
    printf("Invalid parameters to request by the appl\n");
#endif
    return 0;
  }
  bzero((char *) &unspecified_addr, sizeof(unspecified_addr));
  header.error      = SUCCESS;
  header.msg_type   = DHCP_REQUEST;
  header.pid        = htons(getpid());
  header.request_no = htonl(request_no++);

  iov[i].iov_base   = (char *) &header;
  iov[i].iov_len    = sizeof(header);
  len              += iov[i++].iov_len;
  
  req_p.reserved_u_c_r_t = 0;

  if(T) {
    req_p.transaction_id = htons(transaction_id);
    SET_BITS_LOW(req_p.reserved_u_c_r_t, 1, 1);
  }
  else
    req_p.transaction_id = htons(0);

  SET_BITS_MIDDLE(req_p.reserved_u_c_r_t, R, 2, 1);
  SET_BITS_MIDDLE(req_p.reserved_u_c_r_t, C, 3, 1);

  if(unicast_addr && !IN6_IS_ADDR_UNSPECIFIED(unicast_addr))
    SET_BITS_MIDDLE(req_p.reserved_u_c_r_t, 1, 4, 1);
  
  req_p.reserved_u_c_r_t = htons(req_p.reserved_u_c_r_t);

  iov[i].iov_base   = (char *) &req_p;
  iov[i].iov_len    = sizeof(req_p);
  len              += iov[i++].iov_len;

  if(GET_BITS_MIDDLE(req_p.reserved_u_c_r_t, 4, 1) /* U bit of request */) {
    iov[i].iov_base = (char *) unicast_addr;
    iov[i].iov_len  = sizeof(*unicast_addr);
    len            += iov[i++].iov_len;
  }
  
  if(!relay_addr)
    iov[i].iov_base = (char *) &unspecified_addr;
  else
    iov[i].iov_base = (char *) relay_addr;
  iov[i].iov_len    = sizeof(struct in6_addr);
  len              += iov[i++].iov_len;

  if(!server_addr || IN6_IS_ADDR_UNSPECIFIED(server_addr)) {
    printf("Application sent NULL or unspecified server address to "
	   "library function request. Returning...\n");
    return 0;
  }
  

  iov[i].iov_base   = (char *) server_addr;
  iov[i].iov_len    = sizeof(*server_addr);
  len              += iov[i++].iov_len;

  /* Attach extensions           */
  /* First IP address extensions */
  for(; ip_ext_head; ip_ext_head = ip_ext_head->next) {
    t_tls = (TYPE_LEN_SETTINGS *) malloc(sizeof(TYPE_LEN_SETTINGS));
    if(!t_tls)
      errx(1, "Insufficient memory\n");
    bzero((char *) t_tls, sizeof(*t_tls));
    t_tls->type     = htons(IP_ADDRESS_EXTENSION);
    t_tls->length   = sizeof(t_tls->settings);
    t_tls->settings = 0;

    iov[i].iov_base   = (char *) t_tls;
    iov[i].iov_len    = sizeof(t_tls->type) + sizeof(t_tls->length) + sizeof(t_tls->settings);
    len              += iov[i++].iov_len;

    SET_BITS_HIGH(t_tls->settings, (ip_ext_head->s).status, 8, 32);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).C, 24, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).I, 23, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).L, 22, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).Q, 21, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).A, 20, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).P, 19, 1);
    SET_BITS_MIDDLE(t_tls->settings, (ip_ext_head->s).scope, 10, 3);
    SET_BITS_LOW(t_tls->settings, (ip_ext_head->s).prefix_len, 7);
    
    if((ip_ext_head->s).C) {
      iov[i].iov_base = (char *) &ip_ext_head->ip_addr;
      iov[i].iov_len  = sizeof(ip_ext_head->ip_addr);
      len            += iov[i].iov_len;
      t_tls->length  += iov[i++].iov_len;
    }
    if((ip_ext_head->s).L) {
      ip_ext_head->preferred_lifetime = htonl(ip_ext_head->preferred_lifetime);
      ip_ext_head->valid_lifetime     = htonl(ip_ext_head->valid_lifetime);
      iov[i].iov_base = (char *) &ip_ext_head->preferred_lifetime;
      iov[i].iov_len  = sizeof(ip_ext_head->preferred_lifetime);
      len            += iov[i].iov_len;
      t_tls->length  += iov[i++].iov_len;

      iov[i].iov_base = (char *) &ip_ext_head->valid_lifetime;
      iov[i].iov_len  = sizeof(ip_ext_head->valid_lifetime);
      len            += iov[i].iov_len;
      t_tls->length  += iov[i++].iov_len;
    }

    if(ip_ext_head->dns_name[0] != '\0') {
      /* dns name is present */
      iov[i].iov_base = (char *) ip_ext_head->dns_name;
      iov[i].iov_len  = strlen(ip_ext_head->dns_name);
      len            += iov[i].iov_len;
      t_tls->length  += iov[i++].iov_len;
    }
    t_tls->length     = htons(t_tls->length);
    t_tls->settings   = htonl(t_tls->settings);
    
    *temp_tls = t_tls;
    temp_tls  = &(*temp_tls)->next;
  }

  /* Extension request extension */
  for(count = 0; count < no_of_ere; count++) {
    ere[count] = htons(ere[count]);
  }

  if(no_of_ere > 0) {
    ere_tl.type       = htons(EXTENSION_REQUEST_EXTENSION);
    ere_tl.length     = htons(no_of_ere * sizeof(*ere));
    iov[i].iov_base   = (char *) &ere_tl;
    iov[i].iov_len    = sizeof(ere_tl);
    len              += iov[i++].iov_len;

    iov[i].iov_base   = (char *) ere;
    iov[i].iov_len    = no_of_ere * sizeof(*ere);
    len              += iov[i++].iov_len;
  }
  /* Remaining extensions        */
  if(remaining_ext_length > 0) {
    iov[i].iov_base   = (char *) remaining_ext;
    iov[i].iov_len    = remaining_ext_length;
    len              += iov[i++].iov_len;
  }

  if(writev(sockfd, &iov[0], i) != len){
    printf("Error in sending request message to client daemon. Returning...\n");
    return 0;
  }

  printf("Sending:< ");
  {
    int count;
    char *c;

    for(count = 0; count < i; count++) {
      for(c = iov[count].iov_base; iov[count].iov_len > 0; c++, iov[count].iov_len--)
	printf("%02x ", (uint8_t)*c);
      printf("\n");
    }
  }
  printf(">\n");
  /* free up the list of type length and settings we made */
  while(tls_head) {
    t_tls = tls_head;
    tls_head = tls_head->next;
    free(t_tls);
  }
#ifdef DEBUG
  printf("Waiting for reply to request msg\n");
#endif

  n = recv(sockfd, recv_msg, MAX_MSG_SIZE, 0);

  if(n < 0){
    printf("Error in recv. Returning...\n");
    return 0;
  }
  if(n == 0){
    printf("Connection closed by client daemon. Returning...\n");
    return 0;
  }

#ifdef DEBUG
  printf("Reply to request-> ");
  print_packet(recv_msg, n);
  printf("\n");

  memcpy(&header, recv_msg, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  printf("Reply from client daemon\n\n");
  printf("msg_type   = %u\n", header.msg_type);
  printf("pid        = %u\n", header.pid);
  printf("request_no = %u\n", header.request_no);
#endif

  return 1;
}
