# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>

# define MAX_TOKENS_PER_LINE 8
# define MAX_TOKEN_LEN       256

typedef struct iaddr{
  struct in6_addr  ip_addr;
  struct iaddr    *next;
  uint8_t          prefix_len;
}IADDR;

typedef struct interface{
  struct interface *next;
  uint32_t          index;
  char              name[IFNAMSIZ];
  struct in6_addr  *link_local_addr;
  IADDR            *addr_list_head;
}INTERFACE;

char *INTERFACE_CONF_FILE;
char line[MAX_TOKENS_PER_LINE][MAX_TOKEN_LEN];
INTERFACE *in_list_head;
uint line_no = 0;
char name[64];

int read_line(FILE *fp){
  int i = 0, n;
  char c;

  while(1){
    fscanf(fp, "%*[ ]");
    n = fscanf(fp, "%[^\n ]", line[i]);
    if(n <= 0){
      c = fgetc(fp);
      if(c == '\n'){
	line_no++;
	break;
      }
      if(c == EOF)
	return -1;
    }
    i++;
  }
  return i;
}

int read_useful_line(FILE *fp){
  int i;

  while(1){
    i = read_line(fp);
    if(i == -1)
      return i;
    if(i != 0 && line[0][0] != '#')
      return i;
  }
}

void print_line(int i){
  int k;

  printf("%d\n", i);
  for(k = 0; k < i; k++)
    printf("%s ", line[k]);
  printf("\n");
}

uint8_t get_and_delete_prefix_len(char addr[]){
  int i;

  for(i = strlen(addr) - 1; i > 0; i--)
    if(addr[i] == '/')
      break;

  if(i == 0)
    errx(1, "Address doesn't contain prefix. Successfully read %u lines\n", line_no);

  addr[i] = '\0';
  return atoi(addr + i + 1);
}

INTERFACE *read_interface(FILE *fp){
  int         i;
  INTERFACE  *in;
  IADDR     **t;
  
  i = read_useful_line(fp);
  switch(i){

  case -1:
    /* EOF reached */
    return NULL;
    break;
    
  case 2:
    if(line[1][strlen(line[1]) - 1] != '{')
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   INTERFACE_CONF_FILE, line_no);

    line[1][strlen(line[1]) - 1] = '\0';
    break;
    
  case 3:
    break;
    
  default:
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 INTERFACE_CONF_FILE, line_no);
    break;
  }

  if(strcmp(line[0], "interface") != 0)
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 INTERFACE_CONF_FILE, line_no);
  
  in = (INTERFACE *) malloc(sizeof(INTERFACE));
  if(!in)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) in, sizeof(INTERFACE));
  strcpy(in->name, line[1]);
  t  = &in->addr_list_head;
  
  while(1){
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      //return in;
      errx(2, "Unexpected end of file\n");
      break;
      
    case 1:
      if(strcmp(line[0], "}") == 0){
	//printf("End of interface\n\n");
	return in;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     INTERFACE_CONF_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "link-local-address") == 0){
	
	if(in->link_local_addr){
	  warnx("link-local-address mentioned more than once\n");
	}
	else{
	  in->link_local_addr = (struct in6_addr *)malloc(sizeof(struct in6_addr));
	  if(!in->link_local_addr)
	    errx(1, "Insufficient memory\n");
	}
	
	if(inet_pton(AF_INET6, line[1], in->link_local_addr) <= 0)
	  errx(1, "Error in converting link local addr to binary\n");
	
      }
      else if(strcmp(line[0], "index") == 0){
	if(in->index != 0)
	  warnx("index specified more than once\n");
	in->index = atoi(line[1]);
      }
      else if(strcmp(line[0], "ip-address-list") == 0){
	if(in->addr_list_head)
	  warnx("ip-address-list mentioned more than once. Will concatenate lists\n");

	*t = (IADDR *) malloc(sizeof(IADDR));
	if(!(*t))
	  errx(1, "Insufficient memory\n");
	
	/* line[1] contains address in format address/prefix-len */
	(*t)->prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(*t)->ip_addr) <= 0)
	  errx(1, "Error in converting link local addr to binary\n");

	t = &(*t)->next;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     INTERFACE_CONF_FILE, line_no);
       
      break;
      
    default:
      if(strcmp(line[0], "ip-address-list") == 0){
	int k;

	if(in->addr_list_head)
	  warnx("ip-address-list mentioned more than once. Will concatenate lists\n");
	
	for(k = 1; k < i; k++){
	  
	  *t = (IADDR *) malloc(sizeof(IADDR));
	  if(!(*t))
	    errx(1, "Insufficient memory\n");
	
	  (*t)->prefix_len = get_and_delete_prefix_len(line[k]);
	  if(inet_pton(AF_INET6, line[k], &(*t)->ip_addr) <= 0)
	    errx(1, "Error in converting link local addr to binary\n");
	  
	  t = &(*t)->next;
	}
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     INTERFACE_CONF_FILE, line_no);
      break;
    }
  }
}

void print_interface(INTERFACE *interface){
  IADDR *temp_iaddr;

  printf("name: %s\n", interface->name);
  printf("index: %d\n", interface->index);
  if(interface->link_local_addr)
    printf("link-local-address: %s\n", 
	   inet_ntop(AF_INET6, interface->link_local_addr, name, 64));
  printf("ip-address-list: ");
  for(temp_iaddr = interface->addr_list_head; temp_iaddr; temp_iaddr = temp_iaddr->next){
    printf("%s/%u ",
	   inet_ntop(AF_INET6, &temp_iaddr->ip_addr, name, 64),
	   temp_iaddr->prefix_len);
  }
  printf("\n");

}

void parse_interface_file(){
  FILE *fp;
  int i, k;
  INTERFACE **interface = &in_list_head;

  fp = fopen(INTERFACE_CONF_FILE, "r");
  if(!fp)
    errx(1, "Error opening %s. Exiting...\n", INTERFACE_CONF_FILE);

  while(1){
    *interface = read_interface(fp);
    if(!(*interface))
      break;
    print_interface(*interface);
    interface = &(*interface)->next;
    
  }
}

main(int argc, char *argv[]){
  if(argc != 2){
    printf("Incorrect number of arguments to main\n");
    exit(0);
  }

  INTERFACE_CONF_FILE = argv[1];
  parse_interface_file();
}
